from nilmdb.utils.printf import *
import sys
import inspect
import functools

def must_close(errorfile = sys.stderr, wrap_verify = False):
    """Class decorator that warns on 'errorfile' at deletion time if
    the class's close() member wasn't called.

    If 'wrap_verify' is True, every class method is wrapped with a
    verifier that will raise AssertionError if the .close() method has
    already been called."""
    def decorator(cls):
        def dummy(*args, **kwargs):
            pass
        if "__init__" not in cls.__dict__:
            cls.__init__ = dummy
        if "__del__" not in cls.__dict__:
            cls.__del__ = dummy
        if "close" not in cls.__dict__:
            cls.close = dummy

        orig_init = cls.__init__
        orig_del = cls.__del__
        orig_close = cls.close

        def __init__(self, *args, **kwargs):
            ret = orig_init(self, *args, **kwargs)
            self.__dict__["_must_close"] = True
            self.__dict__["_must_close_initialized"] = True
            return ret

        def __del__(self):
            if "_must_close" in self.__dict__:
                fprintf(errorfile, "error: %s.close() wasn't called!\n",
                        self.__class__.__name__)
            return orig_del(self)

        def close(self, *args, **kwargs):
            del self._must_close
            return orig_close(self)

        cls.__init__ = __init__
        cls.__del__ = __del__
        cls.close = close

        # Optionally wrap all other functions
        def make_verifier(func):
            @functools.wraps(func)
            def verifier(self, *args, **kwargs):
                if ("_must_close" not in self.__dict__ and
                    "_must_close_initialized" in self.__dict__):
                    raise AssertionError("called " + str(func) + " after close")
                return func(self, *args, **kwargs)
            return verifier
        if wrap_verify:
            for (name, method) in inspect.getmembers(cls, inspect.ismethod):
                # Skip class methods
                if method.__self__ is not None:
                    continue
                # Skip some methods
                if name in [ "__del__", "__init__" ]:
                    continue
                # Set up wrapper
                setattr(cls, name, make_verifier(method))

        return cls
    return decorator
